#ifndef _event_
#define _event_ 1

#include "vector.h"
#include <iostream>
// using namespace std;
#include <cstdio>
#include <cmath>
#include <clip.bin.h>

class Event {
   float x, y;
 public:
   Event() { }
   Event( double x, double y ) { this->x = (float) x; this->y = (float) y; }

   double getX() const { return x; }
   double getY() const { return y; }

   void setX(double x) { this->x = (float) x; }
   void setY(double y) { this->y = (float) y; }

   double norm() const { return sqrt( x*x + y*y ); }

   virtual void report() const 
      { std::cout << "Event: ( " << x << ", " << y << " )\n"; }
};

typedef Event *EventPointer;

double distance( const Event&, const Event& );

inline
Event operator - ( const Event& a ) { return Event( -a.getX(), -a.getY() ); }

inline
Event operator + ( const Event& a, const Event& b ) {
   return Event( a.getX()+b.getX(), a.getY()+b.getY() );
}

inline
Event operator - (const Event& a, const Event& b )
{
   return Event( a.getX()-b.getX(), a.getY()-b.getY() );
}

inline
bool operator == ( const Event& a, const Event& b ) 
{
   return ( a.getX() == b.getX() ) && ( a.getY() == b.getY() ); 
}

inline
bool operator != ( const Event& a, const Event& b ) 
{
   return ( a.getX() != b.getX() ) || ( a.getY() != b.getY() ); 
}

std::ostream& operator << ( std::ostream&, Event );

/*

	Object		object geometry for translation models:
			- typically maintains template object 
			- rules for geometry
				e.g. for finding all pixels in object

	Pattern		pattern of objects:
			- typically maintains list of object centres
			- could also maintain neighbourhood relations
			- rules for adding/deleting objects in pattern
*/

class Object : public Event {
   int colour;
 public:
   Object( double xc, double yc, int cc ) : Event(xc, yc) { colour = cc; }
   int getColour() const { return colour; }
   void setColour( int c ) { colour = c; }

   virtual sequence shape() const = 0;
        // the pixels represented by the object
        // Beware: not clipped to fit inside some data image!
   virtual int overlap( const Object * ) const = 0;
   virtual int contains( point p ) const = 0;
        // Boolean functions to test whether two objects 
        // overlap, respectively p is contained in the object
   virtual void report() const = 0;
};

class VideoObject : public Object {
 public:
   VideoObject *parent;
   VideoObject *child;
   // int frame;

   VideoObject( VideoObject *cp, VideoObject *cc, double xc, double yc, int ccol ) :
      Object(xc, yc, ccol) { 
         parent = cp; child = cc; 
         parent->child = this; child->parent = this; }
   VideoObject( double xc, double yc, int ccol ) :
      Object(xc, yc, ccol) { parent = NULL; child = NULL; }
};

/*

   Square. Position is center
   Marks are sidelength (float) and colour (grey level)

*/

class Square: public VideoObject {
   float side;
   sequence pixels;
 public:
   Square( VideoObject *cp, VideoObject *cc, double xc, double yc, double sc, int ccol );

   Square( double xc, double yc, double sc, int ccol );
   double getSide() const { return side; }
   void setSide( double s );

   int overlap( const Object * ) const;
   sequence shape() const;
   int contains( point p ) const;

   void report() const;
};

class Ellips: public VideoObject {
   float horaxis;  /* length of horizontal axis */
   float vertaxis; /* length of vertical axis   */
   float ort;      /* orientation in radials    */
   sequence pixels;
 public:
   Ellips( VideoObject *cp, VideoObject *cc, double xc, double yc, 
           double horc, double vertc, double phic, int ccol );
   Ellips( double xc, double yc, 
           double horc, double vertc, double phic, int ccol );

   double getOrientation() const { return ort; }
   void setOrientation( double phi );
   double getMajor() const { return horaxis; }
   void setMajor( double m );
   double getMinor() const { return vertaxis; }
   void setMinor( double m );

   int overlap( const Object * ) const;
   sequence shape() const;
   int contains( point p ) const;

   void report() const;
};

#endif

