#ifndef _prior_
#define _prior_ 1

#include "pattern.h"
#include "random.h"

/*
	Locally stable sequential Markov priors 
	on rectangular windows

	member functions for computation of and 
	upper bound on the conditional intensity

*/

class Prior {
   float width, height; // rectangle in which model is defined
 protected:
   float K;             // local stability bound
   FILE *par_file;      // for the parameters 
   FILE *stats_file;    // for the statistics
   FILE *pat_file;      // for a pattern
 public:
   Prior( double cK, double cw, double ch);
   virtual ~Prior() { }

   virtual double cond_intens( const Pattern& p, const Event* e, int i ) const = 0;
   virtual double logCond_intens( const Pattern& p, const Event* e, int i ) const 
    { return( log( cond_intens( p, e, i ) ) ); }
   virtual int monotone() const = 0;
      // 0 for repulsive, 1 for attractive, 2 otherwise

   double getWidth() const { return width; }
   double getHeight() const { return height; }
   void setWidth( double cw ) { width = (float) cw; }
   void setHeight( double ch ) { height = (float) ch; }

   double ubnd() const { return K; }
     // NB: upper bound of f(x + u)/f(x), no factor for size!
   virtual void report() const = 0;

   void printInit( char *name_par, char *name_stats, char *name_pat );
   void printClose();

   virtual void printParameters(); 
      // by default the ubnd 
   virtual void printSample( const Pattern& p );
   virtual void printStatistics( const Pattern& p );
      // by default the total number of points
   int getTotal( const Pattern& p ) const;
   
   virtual Event *newEvent() const;
   virtual double newLocal( Event *e, Event *f, double d ) const
     { f = e; return(1.0); }
      // return value is relative area of local neighbourhoods
};

class Poisson : public Prior {
  public:
   Poisson( double cK, double cw, double ch) 
      : Prior( cK, cw, ch ) { };
   void setParameters( double cb ) { K = cb; }
   double cond_intens( const Pattern& p, const Event *e, int i ) const
      { return( K / ( 1.0 + p.getSize() ) ); }
   int monotone() const { return 2; }
   void report() const { printf("Poisson process of rate %f\n", K); } 
};
#endif
