#include "softcore.h"

Softcore::Softcore( double cb, double cg, double cm, double cw, double ch ) :
   Prior( cb, cw, ch )
{
   if( cg < 0.00001 ) { 
     printf("Softcore: gamma must be positive\n");
     exit(1);
   }
   if( cg > 1.0 ) { 
     printf("Softcore: gamma must be in unit interval\n");
     exit(1);
   }

   if( cm < 0.0001 ) { 
     printf("Softcore: mu must be positive\n");
     exit(1);
   }

   this->gamma = cg;
   this->mu = cm;
}

double Softcore::cond_intens( const Pattern& p, const Event* e, int i ) const
{
   int sz = p.getSize();
   if( i > sz || i < 0 ) { 
      printf("Softcore: cond_intens range out of bound\n");
      exit(1);
   }

   MarkedEvent *le = (MarkedEvent *) e;
   MarkedEvent *f;
 
   int expsum = 0;

   for( int j=0; j<i; j++ ) {
      f = (MarkedEvent *) p.getEvent(j);
      if( le->getDistance(f) < 1.0001*f->getMark() ) expsum++;
   }
   for( int j=i; j<sz; j++ ) {
      f = (MarkedEvent *) p.getEvent(j);
      if( le->getDistance(f) < 1.0001*le->getMark() ) expsum++;
   }

   return( ( ubnd() / (1.0 + sz ) ) * exp( log(gamma) * expsum ) );
}

int
Softcore::monotone() const {
   return 2; 
}

void 
Softcore::setParameters( double cb, double cg, double cm )
{
   K = cb;

   if( cg < 0.0 ) { 
     printf("Softcore: gamma must be in unit interval\n");
     exit(1);
   }
   if( cg > 1.0 ) { 
     printf("Softcore: gamma must be in unit interval\n");
     exit(1);
   }

   if( cm < 0.0 ) { 
     printf("Softcore: mu must be positive\n");
     exit(1);
   }

   this->gamma = cg;
   this->mu = cm;
}

void 
Softcore::report() const
{
   printf("Poisson-exponential sequential soft core model with");
   printf(" intensity parameter %.3f, interaction %.3f, and mean radius %.3f\n", 
          K, gamma, 1/mu);
}

void 
Softcore::printParameters() 
{
   fprintf(par_file, "%f %f %f\n", ubnd(), getGamma(), getMu());
   fflush(par_file);
}

void 
Softcore::printStatistics( const Pattern& p )
{
   fprintf(stats_file, "%d %d\n", getTotal(p), getInteractingPairs(p));
   fflush(stats_file);
}

void 
Softcore::printSample( const Pattern& p )
{
   MarkedEvent *le;
   int sz = p.getSize();
   if( sz > 0 ) {
      for( int i=0; i<sz; i++ ) {
      le = (MarkedEvent*) p.getEvent(i);
      fprintf(pat_file, "%f %f %f \n",
         le->getX(), le->getY(), le->getMark() );
      }
   }
   fflush(pat_file);
}

int 
Softcore::getInteractingPairs( const Pattern& p ) const
{
   // sum_{i<j} 1{ || x_i - x_j || <= m_i }

   int sz = p.getSize();
   int pairs = 0;
   MarkedEvent *e;
 
   for( int i=0; i<sz-1; i++ )
         e = (MarkedEvent*) p.getEvent(i);
      for( int j= i+1; j<sz; j++ ) {
         if( e->getDistance( p.getEvent(j) ) < 1.0001*e->getMark() ) pairs++;
   }
 
   return pairs;
}
 
EventPointer 
Softcore::newEvent() const
{ 
   Point p( getWidth()*whran.uniform(), getHeight()*whran.uniform() );
   return new MarkedEvent( p, whran.exponential(mu) );
}

