#include "strauss.h"

SquareStrauss::SquareStrauss( double clb, double clg, double cw, double ch, 
      int cng, int cbg, double cmax, double cmin, double cv  ) : Prior( exp(clb), cw, ch )
{
   this->loggamma = (float) clg;
   this->logbeta = (float) clb;
   this->ngrey = cng;
   this->bgcolour = cbg;
   this->maxside = (float) cmax;
   this->minside = (float) cmin;
   this->var = (float) cv;
}

int
SquareStrauss::monotone() const {
   return 0; 
}

void 
SquareStrauss::setParameters( double clb, double clg )
{
   K = exp(clb);

   this->loggamma = (float)clg;
   this->logbeta = (float) clb;
}

void 
SquareStrauss::report() const
{
   printf("Sequential Strauss process for overlapping squares with");
   printf(" log intensity parameter %.3f, and log interaction %.3f\n", logbeta, loggamma);
}

void 
SquareStrauss::printParameters() 
{
   fprintf(par_file, "%f %f\n", getLogBeta(), getLogGamma());
   fflush(par_file);
}

void 
SquareStrauss::printStatistics( const Pattern& p )
{
   fprintf(stats_file, "%d %d\n", getTotal(p), getInteractingPairs(p));
   fflush(stats_file);
}

void 
SquareStrauss::printSample( const Pattern& p )
{
   Square *le;
   int sz = p.getSize();
   if( sz > 0 ) {
      for( int i=0; i<sz; i++ ) {
        le = (Square*) p.getEvent(i);
        fprintf(pat_file, "%f %f %f %d\n",
          le->getX(), le->getY(), le->getSide(), le->getColour() );
      }
   }
   fflush(pat_file);
}

int 
SquareStrauss::getInteractingPairs( const Pattern& p ) const
{
   int sz = p.getSize();
   int pairs = 0;
   Square *e;
   Square *f;
 
   for( int i=0; i<sz-1; i++ ) {
      e = (Square*) p.getEvent(i);
      for( int j= i+1; j<sz; j++ ) {
         f = (Square*) p.getEvent(j);
         if( e->overlap(f) ) pairs++;
      }
   }
 
   return pairs;
}

double SquareStrauss::logCond_intens( const Pattern& p, const Event* e, int i ) const
{
   int sz = p.getSize();
   if( i > sz || i < 0 ) { 
      printf("SquareStrauss: cond_intens range out of bound: i=%d, sz=%d\n", i, sz);
      exit(1);
   }

   Square *le = (Square *) e;
   Square *f;

   int expsum = 0;
   for( int j=0; j<sz; j++ ) {
      f = (Square *) p.getEvent(j);
      if( le->overlap(f) ) expsum++;
   }

   return( ( loggamma * expsum ) + logbeta - log( 1.0 + sz ) );
}

EventPointer 
SquareStrauss::newEvent() const
{ 
   return new Square( getWidth()*whrg.uniform(), getHeight()*whrg.uniform(),
       minside + ((maxside - minside)*whrg.uniform()), 
       whrg.discrete_uniform(0, ngrey - 1) );
}

double 
SquareStrauss::logRho( const Object *current, const Object *next ) const
{
   double d = ( *current - *next ).norm(); 
   d = ( ( d*d ) / var );
   double c = current->getColour() - next->getColour();
   double s = ((Square* )current)->getSide() - ((Square *)next)->getSide();

   return( - d - fabs( c / (1.0*ngrey) ) - fabs(s) );

}

Object * 
SquareStrauss::from_kernel( const Object *obj ) const
{
   if( ( obj->getX() < - 3*sqrt(var) ) || 
       ( obj->getX() > getWidth() + 3*sqrt(var) ) ) { 
          std::cout << "Cannot sample from truncated kernel\n";
          obj->report();
          exit(1);
   }
   if( ( obj->getY() < - 3*sqrt(var) ) || 
       ( obj->getY() > getHeight() + 3*sqrt(var) ) ) { 
          std::cout << "Cannot sample from truncated kernel\n";
          obj->report();
          exit(1);
   }

   double x, y;
   while(1) {
      x = whrg.normal(obj->getX(), sqrt(var));
      y = whrg.normal(obj->getY(), sqrt(var));
      if( ( x > 0 ) && ( x < getWidth() ) ) {
         if( ( y > 0 ) && ( y < getHeight() ) ) break;
      }
   }

   return new Square( x, y, ((Square *) obj)->getSide(), 
                      ((Square *) obj)->getColour() );
}

double 
SquareStrauss::logKernel( const Object *current, const Object *next ) const
{
   double d = ( *current - *next ).norm();
   double e = log( 2*M_PI ) + log( var );
   return( - (d*d)/(2*var) - (0.5*e) );
}

bool
SquareStrauss::split( const Object *ob, Object *s1, Object *s2 ) const
{
   double u = maxside * whrg.uniform();
   double v = maxside * whrg.uniform();

   s1->setX( ob->getX() + u - ( maxside / 2.0 ) );
   s1->setY( ob->getY() + v - ( maxside / 2.0 ) );
   ((Square *)s1)->setSide( minside +  ( (maxside - minside) * whrg.uniform() ) );

   s2->setX( ob->getX() - u + ( minside / 2.0 ) );
   s2->setY( ob->getY() - v + ( minside / 2.0 ) );
   ((Square *)s2)->setSide( minside +  ( (maxside - minside) * whrg.uniform() ) );

   if( whrg.uniform() < 0.5 ) {
      s1->setColour( whrg.discrete_uniform( 0, ob->getColour() ) );
      s2->setColour( whrg.discrete_uniform( ob->getColour(), ngrey - 1 ) );
   }
   else {
      s2->setColour( whrg.discrete_uniform( 0, ob->getColour() ) );
      s1->setColour( whrg.discrete_uniform( ob->getColour(), ngrey - 1 ) );
   }

   if( s1->getX() < 0 || s1->getX() > getWidth() ) return false;
   if( s1->getY() < 0 || s1->getY() > getHeight() ) return false;
   if( s2->getX() < 0 || s2->getX() > getWidth() ) return false;
   if( s2->getY() < 0 || s2->getY() > getHeight() ) return false;
   return true;
}

bool
SquareStrauss::merge( const Object *ob1, const Object *ob2, Object *mob ) const
{

   double x1 = ob1->getX() + ( ((Square *)ob1)->getSide() / 2.0 );
   double y1 = ob1->getY() + ( ((Square *)ob1)->getSide() / 2.0 );
   
   double x2 = ob2->getX() + ( ((Square *)ob2)->getSide() / 2.0 );
   double y2 = ob2->getY() + ( ((Square *)ob2)->getSide() / 2.0 );
 
   double rho = fabs( x1 - x2 );
   if( fabs( y1 - y2 ) > rho ) rho = fabs( y1 - y2 );

   if( rho > maxside ) return false;

   mob->setX( ( ob1->getX() + ob2->getX() ) / 2.0 );
   mob->setY( ( ob1->getY() + ob2->getY() ) / 2.0 );
   ((Square *)mob)->setSide( minside +  ( (maxside - minside) * whrg.uniform() ) );

   int minc = ob1->getColour();
   if( ob2->getColour() < minc ) minc = ob2->getColour();
   int maxc = ob1->getColour();
   if( ob2->getColour() > maxc ) maxc = ob2->getColour();
   mob->setColour( whrg.discrete_uniform( minc, maxc ) );
   
   return true;
}

double 
SquareStrauss::logSplitrate( const Object *ob, 
                             const Object *s1, const Object *s2 ) const
{
   double val = log(4.0) + 2.0 * log(maxside) + log(maxside-minside);

   int minc = s1->getColour();
   if( s2->getColour() < minc ) minc = s2->getColour();
   int maxc = s1->getColour();
   if( s2->getColour() > maxc ) maxc = s2->getColour();
   val -= log( maxc - minc + 1.0 );

   val += log(2.0) + log(ngrey - ob->getColour()) + log(ob->getColour() + 1.0);

   return( val );
}

double 
SquareStrauss::logMergerate( const Object *s1, const Object *s2, 
                             const Object *ob ) const
{
   return ( - logSplitrate( ob, s1, s2 ) );
}

double 
change_location( Event *sq, Event *tsq, 
                 double width, double height, double alpha )
{
   double dx = alpha * width;
   double dy = alpha * height;

   double x1 = ( sq->getX() - dx > 0 ? sq->getX() - dx : 0 );
   double x2 = ( sq->getX() + dx < width ? sq->getX() + dx : width );
   double y1 = ( sq->getY() - dy > 0 ? sq->getY() - dy : 0 );
   double y2 = ( sq->getY() + dy < height ? sq->getY() + dy : height );

   double numerator = ( x2 - x1 ) * ( y2 - y1 );

   double tx = x1 + ( x2 - x1 ) * whrg.uniform();
   double ty = y1 + ( y2 - y1 ) * whrg.uniform();
   tsq->setX( tx );
   tsq->setY( ty );

   x1 = ( tx - dx > 0.0 ? tx - dx : 0.0 );
   x2 = ( tx + dx < width ? tx + dx : width );
   y1 = ( ty - dy > 0.0 ? ty - dy : 0.0 );
   y2 = ( ty + dy < height ? ty + dy : height );

   double denominator = ( x2 - x1 ) * ( y2 - y1 );
   return( numerator/ denominator );
}

double 
change_colour( Object *sq, Object *tsq, int ngrey, double alpha )
{
   if( ngrey < 2 ) { std::cout << "No colour to change\n"; exit(1); }

   int ng = (int) round( ngrey * alpha );
   int ng1 = ( sq->getColour() - ng > 0 ?  sq->getColour() - ng : 0 );
   int ng2 = ( sq->getColour() + ng < ngrey ? sq->getColour() + ng : ngrey-1 );

   double numerator = ( ng1 == ng2 ? 1 : ng2 - ng1 + 1 );

   int tc = whrg.discrete_uniform( ng1, ng2 );
   tsq->setColour( tc );


   ng1 = ( tc - ng > 0 ? tc - ng : 0 );
   ng2 = ( tc + ng < ngrey ? tc + ng : ngrey-1 );

   double denominator = ( ng1 == ng2 ? 1.0 : ng2 - ng1 + 1.0 );
   return( numerator/ denominator );
}

double
change_size( Square *sq, Square *tsq, double mins, double maxs, double alpha )
{
   double ds = alpha * ( maxs - mins );
   double s1 = ( sq->getSide() - ds > mins ? sq->getSide() - ds : mins ); 
   double s2 = ( sq->getSide() + ds < maxs ? sq->getSide() + ds : maxs ); 

   double numerator = s2 - s1;

   double ts = s1 + ( s2 - s1 ) * whrg.uniform();
   tsq->setSide( ts );

   s1 = (ts - ds > mins ? ts - ds : mins ); 
   s2 = (ts + ds < maxs ? ts + ds : maxs ); 

   double denominator = s2 - s1;
   return( numerator/ denominator );
}


double SquareStrauss::newLocal( Event *sq, Event *tsq, double alpha ) const
{
   if ( alpha < 0.0 || alpha > 1.0 ) exit(1);

   ((Square *) tsq)->setX( ((Square *)sq)->getX() );
   ((Square *) tsq)->setY( ((Square *)sq)->getY() );
   ((Square *) tsq)->setColour( ((Square *)sq)->getColour() );
   ((Square *) tsq)->setSide( ((Square *)sq)->getSide() );

   ((Square *) tsq)->parent = ((Square *)sq)->parent;
   ((Square *) tsq)->child = ((Square *)sq)->child;

   int ww = whrg.discrete_uniform(1, 3);
   if( ww == 1 ) 
      return change_location( sq, tsq, getWidth(), getHeight(), alpha );

   if( ww == 2 ) 
      return change_colour( (Object *)sq, (Object *)tsq, ngrey, alpha );
    
   if( ww == 3 ) 
      return change_size( (Square *)sq, (Square *)tsq, minside, maxside, alpha );

    std::cout << "newLocal: wrong result\n"; exit(1); 
}

